"""Tools for running the set of simulations used for a pre-night briefing."""

__all__ = [
    "AnomalousOverheadFunc",
    "run_prenights",
    "prenight_sim_cli",
]

import argparse
import bz2
import gzip
import io
import logging
import lzma
import os
import pickle
import typing
from datetime import datetime
from functools import partial
from tempfile import TemporaryFile
from typing import Callable, Optional, Sequence
from warnings import warn

import numpy as np
import numpy.typing as npt
from astropy.time import Time
from matplotlib.pylab import Generator

from rubin_scheduler.scheduler.example import example_scheduler
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.schedulers.core_scheduler import CoreScheduler
from rubin_scheduler.scheduler.utils import SchemaConverter
from rubin_scheduler.sim_archive.sim_archive import drive_sim
from rubin_scheduler.site_models import Almanac

from .make_snapshot import add_make_scheduler_snapshot_args, get_scheduler, save_scheduler

try:
    from rubin_sim.data import get_baseline  # type: ignore
except ModuleNotFoundError:
    get_baseline = partial(warn, "Cannot find default baseline because rubin_sim is not installed.")

DEFAULT_ARCHIVE_URI = "s3://rubin-scheduler-prenight/opsim/"


def _run_sim(
    sim_start_mjd: float,
    archive_uri: str,
    scheduler_io: io.BufferedRandom,
    label: str,
    tags: Sequence[str] = tuple(),
    sim_duration: float = 2,
    anomalous_overhead_func: Optional[Callable] = None,
    opsim_metadata: dict | None = None,
) -> None:
    logging.info(f"Running {label}.")

    scheduler_io.seek(0)
    scheduler = pickle.load(scheduler_io)

    observatory = ModelObservatory(
        nside=scheduler.nside,
        cloud_data="ideal",
        seeing_data="ideal",
        downtimes="ideal",
    )

    drive_sim(
        observatory=observatory,
        scheduler=scheduler,
        archive_uri=archive_uri,
        label=label,
        tags=tags,
        script=__file__,
        sim_start_mjd=sim_start_mjd,
        sim_duration=sim_duration,
        record_rewards=True,
        anomalous_overhead_func=anomalous_overhead_func,
        opsim_metadata=opsim_metadata,
    )


def _mjd_now() -> float:
    # Used instead of just Time.now().mjd to make type checker happy.
    mjd = Time.now().mjd
    assert isinstance(mjd, float)
    return mjd


def _iso8601_now() -> str:
    # Used instead of just Time.now().mjd to make type checker happy.
    now_iso = Time.now().iso
    assert isinstance(now_iso, str)
    return now_iso


def _create_scheduler_io(
    day_obs_mjd: float,
    scheduler_fname: Optional[str] = None,
    scheduler_instance: Optional[CoreScheduler] = None,
    opsim_db: str | None = None,
) -> io.BufferedRandom:
    if scheduler_instance is not None:
        scheduler = scheduler_instance
    elif scheduler_fname is None:
        sample_scheduler = example_scheduler()
        if not isinstance(sample_scheduler, CoreScheduler):
            raise TypeError()

        scheduler = sample_scheduler

    else:
        opener: typing.Callable = open

        if scheduler_fname.endswith(".bz2"):
            opener = bz2.open
        elif scheduler_fname.endswith(".xz"):
            opener = lzma.open
        elif scheduler_fname.endswith(".gz"):
            opener = gzip.open

        with opener(scheduler_fname, "rb") as sched_io:
            scheduler = pickle.load(sched_io)

    scheduler.keep_rewards = True

    if opsim_db is not None:
        last_preexisting_obs_mjd = day_obs_mjd + 0.5
        obs = SchemaConverter().opsim2obs(opsim_db)
        obs = obs[obs["mjd"] < last_preexisting_obs_mjd]
        if len(obs) > 0:
            scheduler.add_observations_array(obs)

    scheduler_io = TemporaryFile()
    pickle.dump(scheduler, scheduler_io)
    scheduler_io.seek(0)
    return scheduler_io


class AnomalousOverheadFunc:
    """Callable to return random overhead.

    Parameters
    ----------
    seed : `int`
        Random number seed.
    slew_scale : `float`
        The scale for the scatter in the slew offest (seconds).
    visit_scale : `float`, optional
        The scale for the scatter in the visit overhead offset (seconds).
        Defaults to 0.0.
    slew_loc : `float`, optional
        The location of the scatter in the slew offest (seconds).
        Defaults to 0.0.
    visit_loc : `float`, optional
        The location of the scatter in the visit offset (seconds).
        Defaults to 0.0.
    """

    def __init__(
        self,
        seed: int,
        slew_scale: float,
        visit_scale: float = 0.0,
        slew_loc: float = 0.0,
        visit_loc: float = 0.0,
    ) -> None:
        self.rng: Generator = np.random.default_rng(seed)
        self.visit_loc: float = visit_loc
        self.visit_scale: float = visit_scale
        self.slew_loc: float = slew_loc
        self.slew_scale: float = slew_scale

    def __call__(self, visittime: float, slewtime: float) -> float:
        """Return a randomized offset for the visit overhead.

        Parameters
        ----------
        visittime : `float`
            The visit time (seconds).
        slewtime : `float`
            The slew time (seconds).

        Returns
        -------
        offset: `float`
            Random offset (seconds).
        """

        slew_overhead: float = slewtime * self.rng.normal(self.slew_loc, self.slew_scale)

        # Slew might be faster that expected, but it will never take negative
        # time.
        if (slewtime + slew_overhead) < 0:
            slew_overhead = 0.0

        visit_overhead: float = visittime * self.rng.normal(self.slew_loc, self.slew_scale)
        # There might be anomalous overhead that makes visits take longer,
        # but faster is unlikely.
        if visit_overhead < 0:
            visit_overhead = 0.0

        return slew_overhead + visit_overhead


def run_prenights(
    day_obs_mjd: float,
    archive_uri: str,
    scheduler_file: Optional[str] = None,
    opsim_db: Optional[str] = None,
    minutes_delays: tuple[float, ...] = (0, 1, 10, 60, 240),
    anomalous_overhead_seeds: tuple[int, ...] = (101, 102, 103, 104, 105),
    sim_nights: int = 2,
    opsim_metadata: dict | None = None,
) -> None:
    """Run the set of scheduler simulations needed to prepare for a night.

    Parameters
    ----------
    day_obs_mjd : `float`
        The starting MJD.
    archive_uri : `str`
        The URI of visits completed before this night.
    scheduler_file : `str`, optional
        File from which to load the scheduler. None defaults to the example
        scheduler in rubin_sim, if it is installed.
        The default is None.
    opsim_db : `str`, optional
        The file name of the visit database for visits preceeding the
        simulation.
        The default is None.
    minutes_delays : `tuple` of `float`
        Delayed starts to be simulated.
    anomalous_overhead_seeds: `tuple` of `int`
        Random number seeds to use for anomalous overhead runs.
    sim_nights: `int`
        Number of nights to simulate. Defaults to 2.
    opsim_metadata: `dict`
        Extra metadata for the archive
    """

    exec_time: str = _iso8601_now()
    scheduler_io: io.BufferedRandom = _create_scheduler_io(
        day_obs_mjd, scheduler_fname=scheduler_file, opsim_db=opsim_db
    )

    # Assign args common to all sims for this execution.
    run_sim = partial(
        _run_sim, archive_uri=archive_uri, scheduler_io=scheduler_io, opsim_metadata=opsim_metadata
    )

    # Find the start of observing for the specified day_obs.
    # Almanac.get_sunset_info does not use day_obs, so just index
    # Almanac.sunsets for what we want directly.
    all_sun_n12_setting: npt.NDArray[np.float_] = Almanac().sunsets["sun_n12_setting"]
    before_first_day_obs: npt.NDArray[np.bool_] = all_sun_n12_setting < day_obs_mjd + 0.5
    after_first_day_obs: npt.NDArray[np.bool_] = all_sun_n12_setting > day_obs_mjd + 1.5
    on_first_day_obs: npt.NDArray[np.bool_] = ~(before_first_day_obs | after_first_day_obs)
    sim_start_mjd: float = all_sun_n12_setting[on_first_day_obs].item()

    all_sun_n12_rising: npt.NDArray[np.float_] = Almanac().sunsets["sun_n12_rising"]
    before_last_day_obs: npt.NDArray[np.bool_] = all_sun_n12_setting < day_obs_mjd + sim_nights + 0.5
    after_last_day_obs: npt.NDArray[np.bool_] = all_sun_n12_setting > day_obs_mjd + sim_nights + 1.5
    on_last_day_obs: npt.NDArray[np.bool_] = ~(before_last_day_obs | after_last_day_obs)
    sim_end_mjd: float = all_sun_n12_rising[on_last_day_obs].item()
    sim_duration: float = sim_end_mjd - sim_start_mjd

    # Begin with an ideal pure model sim.
    completed_run_without_delay: bool = False
    if len(minutes_delays) == 0 or (np.array(minutes_delays) == 0).any():
        run_sim(
            sim_start_mjd,
            label=f"Nominal start and overhead, ideal conditions, run at {exec_time}",
            tags=["ideal", "nominal"],
        )
        completed_run_without_delay = True

    # Delayed start
    for minutes_delay in minutes_delays:
        if completed_run_without_delay and minutes_delay == 0:
            # Did this already.
            continue

        delayed_sim_start_mjd = sim_start_mjd + minutes_delay / (24.0 * 60)
        sim_duration = sim_end_mjd - delayed_sim_start_mjd

        run_sim(
            delayed_sim_start_mjd,
            label=f"Start time delayed by {minutes_delay} minutes,"
            + f" Nominal slew and visit overhead, ideal conditions, run at {exec_time}",
            tags=["ideal", f"delay_{minutes_delay}"],
            sim_duration=sim_duration,
        )

    # Run a few different scatters of visit time
    anomalous_overhead_scale = 0.1
    for anomalous_overhead_seed in anomalous_overhead_seeds:
        anomalous_overhead_func = AnomalousOverheadFunc(anomalous_overhead_seed, anomalous_overhead_scale)
        run_sim(
            sim_start_mjd,
            label=f"Anomalous overhead {anomalous_overhead_seed, anomalous_overhead_scale},"
            + f" Nominal start, ideal conditions, run at {exec_time}",
            tags=["ideal", "anomalous_overhead"],
            anomalous_overhead_func=anomalous_overhead_func,
            sim_duration=sim_duration,
        )


def _parse_dayobs_to_mjd(dayobs: str | float) -> float:
    try:
        day_obs_mjd = Time(dayobs).mjd
    except ValueError:
        try:
            day_obs_mjd = Time(datetime.strptime(str(dayobs), "%Y%m%d"))
        except ValueError:
            day_obs_mjd = Time(dayobs, format="mjd")

    if not isinstance(day_obs_mjd, float):
        raise ValueError

    return day_obs_mjd


def prenight_sim_cli(cli_args: list = []) -> None:

    parser = argparse.ArgumentParser(description="Run prenight simulations")
    default_time = Time(int(_mjd_now() - 0.5), format="mjd")
    parser.add_argument(
        "--dayobs",
        type=str,
        default=default_time.iso,
        help="The day_obs of the night to simulate.",
    )
    parser.add_argument(
        "--archive",
        type=str,
        default=DEFAULT_ARCHIVE_URI,
        help="Archive in which to store simulation results.",
    )
    parser.add_argument("--scheduler", type=str, default=None, help="pickle file of the scheduler to run.")

    # Only pass a default if we have an opsim
    baseline = get_baseline()
    if baseline is not None:
        parser.add_argument(
            "--opsim", default=baseline, type=str, help="Opsim database from which to load previous visits."
        )
    else:
        parser.add_argument("--opsim", type=str, help="Opsim database from which to load previous visits.")

    add_make_scheduler_snapshot_args(parser)

    args = parser.parse_args() if len(cli_args) == 0 else parser.parse_args(cli_args)

    day_obs_mjd = _parse_dayobs_to_mjd(args.dayobs)
    archive_uri = args.archive
    opsim_db = None if args.opsim in ("", "None") else args.opsim

    scheduler_file = args.scheduler
    if args.repo is not None:
        if os.path.exists(scheduler_file):
            raise ValueError(f"File {scheduler_file} already exists!")

        scheduler: CoreScheduler = get_scheduler(args.repo, args.script, args.branch)
        save_scheduler(scheduler, scheduler_file)

        opsim_metadata = {
            "opsim_config_repository": args.repo,
            "opsim_config_script": args.script,
            "opsim_config_branch": args.branch,
        }
    else:
        opsim_metadata = None

    run_prenights(
        day_obs_mjd,
        archive_uri,
        scheduler_file,
        opsim_db,
        minutes_delays=(0, 1, 10, 60),
        anomalous_overhead_seeds=(101, 102),
        opsim_metadata=opsim_metadata,
    )


if __name__ == "__main__":
    prenight_sim_cli()



__all__ = ("FieldSurvey", "FieldAltAzSurvey")

import copy
from functools import cached_property

import numpy as np

from rubin_scheduler.utils import DEFAULT_NSIDE, _approx_alt_az2_ra_dec, _ra_dec2_hpid

from ..detailers import AltAz2RaDecDetailer
from ..features import LastObservation, NObsCount
from ..utils import ObservationArray
from . import BaseSurvey


class FieldSurvey(BaseSurvey):
    """A survey class for running field surveys.

    Parameters
    ----------
    basis_functions : `list` [`rubin_scheduler.scheduler.basis_function`]
        List of basis_function objects
    detailers : `list` [`rubin_scheduler.scheduler.detailer`] objects
        The detailers to apply to the list of observations.
    RA : `float`
        The RA of the field (degrees)
    dec : `float`
        The dec of the field to observe (degrees)
    sequence : `list` [`str`]
        The sequence of observations to take. (specify which filters to use).
    nvisits : `dict` {`str`: `int`}
        Dictionary of the number of visits in each filter.
        Default of None will use a backup sequence of 20 visits per filter.
        Must contain all filters in sequence.
    exptimes : `dict` {`str`: `float`}
        Dictionary of the exposure time for visits in each filter.
        Default of None will use a backup sequence of 38s in u, and
        29.2s in all other bands. Must contain all filters in sequence.
    nexps : dict` {`str`: `int`}
        Dictionary of the number of exposures per visit in each filter.
        Default of None will use a backup sequence of 1 exposure per visit
        in u band, 2 in all other bands. Must contain all filters in sequence.
    ignore_obs : `list` [`str`] or None
        Ignore observations with this string in the `scheduler_note`.
        Will ignore observations which match subsets of the string, as well as
        the entire string. Ignoring 'mysurvey23' will also ignore 'mysurvey2'.
    survey_name : `str` or None.
        The name to give this survey, for debugging and visualization purposes.
        Also propagated to the 'target_name' in the observation.
        The default None will construct a name based on the
        RA/Dec of the field.
    scheduler_note : `str` or None
        The value to include in the scheduler note.
        The scheduler note is for internal, scheduler, use for the purposes of
        identifying observations to ignore or include for a survey or feature.
    readtime : `float`
        Readout time for computing approximate time of observing
        the sequence. (seconds)
    filter_change_time : `float`
        Filter change time, on average. Used for computing approximate
        time for the observing sequence. (seconds)
    nside : `float` or None
        Nside for computing survey basis functions and maps.
        The default of None will use rubin_scheduler.utils.set_default_nside().
    flush_pad : `float`
        How long to hold observations in the queue after they
        were expected to be completed (minutes).
    """

    def __init__(
        self,
        basis_functions,
        RA,
        dec,
        sequence="ugrizy",
        nvisits=None,
        exptimes=None,
        nexps=None,
        ignore_obs=None,
        survey_name=None,
        target_name=None,
        science_program=None,
        observation_reason=None,
        scheduler_note=None,
        readtime=2.4,
        filter_change_time=120.0,
        nside=DEFAULT_NSIDE,
        flush_pad=30.0,
        detailers=None,
    ):
        default_nvisits = {"u": 20, "g": 20, "r": 20, "i": 20, "z": 20, "y": 20}
        default_exptimes = {"u": 38, "g": 29.2, "r": 29.2, "i": 29.2, "z": 29.2, "y": 29.2}
        default_nexps = {"u": 1, "g": 2, "r": 2, "i": 2, "z": 2, "y": 2}

        self.ra = np.radians(RA)
        self.ra_hours = RA / 360.0 * 24.0
        self.dec = np.radians(dec)
        self.ra_deg, self.dec_deg = RA, dec

        self.survey_name = survey_name
        if self.survey_name is None:
            self._generate_survey_name(target_name=target_name)
        # Backfill target name if it wasn't set
        if target_name is None:
            target_name = self.survey_name

        super().__init__(
            nside=nside,
            basis_functions=basis_functions,
            detailers=detailers,
            ignore_obs=ignore_obs,
            survey_name=self.survey_name,
            target_name=target_name,
            science_program=science_program,
            observation_reason=observation_reason,
        )
        # It's useful to save these as class attributes too
        self.target_name = target_name
        self.science_program = science_program
        self.observation_reason = observation_reason

        self.indx = _ra_dec2_hpid(self.nside, self.ra, self.dec)

        # Set all basis function equal.
        self.basis_weights = np.ones(len(basis_functions)) / len(basis_functions)

        self.flush_pad = flush_pad / 60.0 / 24.0  # To days
        self.filter_sequence = []

        self.scheduler_note = scheduler_note
        if self.scheduler_note is None:
            self.scheduler_note = self.survey_name

        # This sets up what a requested "observation" looks like.
        # For sequences, each 'observation' is more than one exposure.
        # When generating actual observations, filters which are not available
        # are not included in the requested sequence.
        if nvisits is None:
            nvisits = default_nvisits
        if exptimes is None:
            exptimes = default_exptimes
        if nexps is None:
            nexps = default_nexps

        # Do a little shuffling if this was not configured quite as expected
        if isinstance(sequence, str):
            if isinstance(nvisits, (float, int)):
                nvisits = dict([(filtername, nvisits) for filtername in sequence])
            if isinstance(exptimes, (float, int)):
                exptimes = dict([(filtername, exptimes) for filtername in sequence])
            if isinstance(nexps, (float, int)):
                nexps = dict([(filtername, nexps) for filtername in sequence])

        if isinstance(sequence, ObservationArray) | isinstance(sequence[0], ObservationArray):
            self.observations = sequence
        else:
            self.observations = []
            for filtername in sequence:
                for j in range(nvisits[filtername]):
                    obs = ObservationArray()
                    obs["filter"] = filtername
                    obs["exptime"] = exptimes[filtername]
                    obs["RA"] = self.ra
                    obs["dec"] = self.dec
                    obs["nexp"] = nexps[filtername]
                    obs["scheduler_note"] = self.scheduler_note
                    self.observations.append(obs)

        # Let's just make this an array for ease of use
        self.observations = np.concatenate(self.observations)
        order = np.argsort(self.observations["filter"])
        self.observations = self.observations[order]

        n_filter_change = np.size(np.unique(self.observations["filter"]))

        # Make an estimate of how long a sequence will take.
        # Assumes no major rotational or spatial
        # dithering slowing things down.
        # Does not account for unavailable filters.
        self.approx_time = (
            np.sum(self.observations["exptime"] + readtime * self.observations["nexp"])
            + filter_change_time * n_filter_change
        )
        # convert to days, for internal approximation in timestep sizes
        self.approx_time /= 3600.0 / 24.0
        # This is the only index in the healpix arrays that will be considered
        self.indx = _ra_dec2_hpid(self.nside, self.ra, self.dec)

        # Tucking this here so we can look at how many observations
        # recorded - both for any note and for this note.
        self.extra_features["ObsRecorded"] = NObsCount(scheduler_note=None)
        self.extra_features["LastObs"] = LastObservation(scheduler_note=None)
        self.extra_features["ObsRecorded_note"] = NObsCount(scheduler_note=self.scheduler_note)
        self.extra_features["LastObs_note"] = LastObservation(scheduler_note=self.scheduler_note)

    def _generate_survey_name(self, target_name=None):
        if target_name is not None:
            self.survey_name = target_name
        else:
            self.survey_name = f"Field {self.ra_deg :.2f} {self.dec_deg :.2f}"

    @cached_property
    def roi_hpid(self):
        hpid = _ra_dec2_hpid(self.nside, self.ra, self.dec)
        return hpid

    def check_continue(self, observation, conditions):
        # feasibility basis functions?
        """
        This method enables external calls to check if a given
        observations that belongs to this survey is
        feasible or not. This is called once a sequence has
        started to make sure it can continue.

        XXX--TODO:  Need to decide if we want to develop check_continue,
        or instead hold the sequence in the survey, and be able to check
        it that way.
        (note that this may depend a lot on how the SchedulerCSC works)
        """
        return True

    def calc_reward_function(self, conditions):
        # only calculates reward at the index for the RA/Dec of the field
        self.reward_checked = True
        if self._check_feasibility(conditions):
            self.reward = 0.0
            for bf, weight in zip(self.basis_functions, self.basis_weights):
                basis_value = bf(conditions, indx=self.indx)
                self.reward += basis_value * weight

            if not np.isscalar(self.reward):
                self.reward = np.sum(self.reward[self.indx])

                if np.any(np.isinf(self.reward)):
                    self.reward = np.inf
        else:
            # If not feasible, negative infinity reward
            self.reward = -np.inf

        return self.reward

    def generate_observations_rough(self, conditions):
        result = []
        if self._check_feasibility(conditions):
            result = copy.deepcopy(self.observations)

            # Set the flush_by
            result["flush_by_mjd"] = conditions.mjd + self.approx_time + self.flush_pad

            # remove filters that are not mounted
            mask = np.isin(result["filter"], conditions.mounted_filters)
            result = result[mask]
            # Put current loaded filter first
            ind1 = np.where(result["filter"] == conditions.current_filter)[0]
            ind2 = np.where(result["filter"] != conditions.current_filter)[0]
            result = result[ind1.tolist() + (ind2.tolist())]

            # convert to list of array.
            final_result = [
                row.reshape(
                    1,
                )
                for row in result
            ]
            result = final_result

        return result

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} survey_name='{self.survey_name}'"
            f", RA={self.ra}, dec={self.dec} at {hex(id(self))}>"
        )


class FieldAltAzSurvey(FieldSurvey):
    """A clone of FieldSurvey that takes alt,az rather than RA,dec.

    Parameters
    ----------
    basis_functions : `list` [`rubin_scheduler.scheduler.basis_function`]
        List of basis_function objects
    detailers : `list` [`rubin_scheduler.scheduler.detailer`] objects
        The detailers to apply to the list of observations.
    az : `float`
        The azimuth of the field (degrees)
    alt : `float`
        The altitude of the field to observe (degrees)
    sequence : `list` [`str`]
        The sequence of observations to take. (specify which filters to use).
    nvisits : `dict` {`str`: `int`}
        Dictionary of the number of visits in each filter.
        Default of None will use a backup sequence of 20 visits per filter.
        Must contain all filters in sequence.
    exptimes : `dict` {`str`: `float`}
        Dictionary of the exposure time for visits in each filter.
        Default of None will use a backup sequence of 38s in u, and
        29.2s in all other bands. Must contain all filters in sequence.
    nexps : dict` {`str`: `int`}
        Dictionary of the number of exposures per visit in each filter.
        Default of None will use a backup sequence of 1 exposure per visit
        in u band, 2 in all other bands. Must contain all filters in sequence.
    ignore_obs : `list` [`str`] or None
        Ignore observations with this string in the `scheduler_note`.
        Will ignore observations which match subsets of the string, as well as
        the entire string. Ignoring 'mysurvey23' will also ignore 'mysurvey2'.
    survey_name : `str` or None.
        The name to give this survey, for debugging and visualization purposes.
        Also propagated to the 'target_name' in the observation.
        The default None will construct a name based on the
        RA/Dec of the field.
    scheduler_note : `str` or None
        The value to include in the scheduler note.
        The scheduler note is for internal, scheduler, use for the purposes of
        identifying observations to ignore or include for a survey or feature.
    readtime : `float`
        Readout time for computing approximate time of observing
        the sequence. (seconds)
    filter_change_time : `float`
        Filter change time, on average. Used for computing approximate
        time for the observing sequence. (seconds)
    nside : `float` or None
        Nside for computing survey basis functions and maps.
        The default of None will use rubin_scheduler.utils.set_default_nside().
    flush_pad : `float`
        How long to hold observations in the queue after they
        were expected to be completed (minutes)."""

    def __init__(
        self,
        basis_functions,
        az,
        alt,
        sequence="ugrizy",
        nvisits=None,
        exptimes=None,
        nexps=None,
        ignore_obs=None,
        survey_name=None,
        target_name=None,
        science_program=None,
        observation_reason=None,
        scheduler_note=None,
        readtime=2.4,
        filter_change_time=120.0,
        nside=DEFAULT_NSIDE,
        flush_pad=30.0,
        detailers=None,
    ):

        if detailers is None:
            detailers = [AltAz2RaDecDetailer()]

        # Check that an AltAz detailer is present
        names_match = ["AltAz2RaDecDetailer" in det.__class__.__name__ for det in detailers]
        if not np.any(names_match):
            ValueError(
                "detailers list does not include a AltAz2RaDecDetailer which is needed for FieldAltAzSurvey"
            )

        self.az = np.radians(az)
        self.alt = np.radians(alt)

        self.alt_deg, self.az_deg = alt, az

        super().__init__(
            basis_functions=basis_functions,
            RA=0.0,
            dec=0.0,
            sequence=sequence,
            nvisits=nvisits,
            exptimes=exptimes,
            nexps=nexps,
            ignore_obs=ignore_obs,
            survey_name=survey_name,
            target_name=target_name,
            science_program=science_program,
            observation_reason=observation_reason,
            scheduler_note=scheduler_note,
            readtime=readtime,
            filter_change_time=filter_change_time,
            nside=nside,
            flush_pad=flush_pad,
            detailers=detailers,
        )
        # Clobber and set to None
        self.ra = None
        self.dec = None

        # Backfill target name if it wasn't set
        if target_name is None:
            target_name = self.survey_name

        # Default self.observations have ra,dec. Swap to alt,az
        self.observations["RA"] = None
        self.observations["dec"] = None
        self.observations["alt"] = self.alt
        self.observations["az"] = self.az

    @property
    def roi_hpid(self):
        ra, dec = _approx_alt_az2_ra_dec([self.alt], [self.az], self.lat, self.lon, self.mjd)
        hpid = _ra_dec2_hpid(self.nside, ra, dec)
        return hpid

    def _generate_survey_name(self, target_name=None):
        if target_name is not None:
            self.survey_name = target_name
        else:
            self.survey_name = f"Field alt,az {self.alt_deg :.2f} {self.az_deg :.2f}"

    def calc_reward_function(self, conditions):
        # Only calculates reward at the roi_hpid for the RA,dec coordinate
        # of the alt,az position at the current time.
        self.reward_checked = True
        self.mjd = conditions.mjd
        self.lat = np.radians(conditions.site.latitude)
        self.lon = np.radians(conditions.site.longitude)
        hpid = self.roi_hpid
        if self._check_feasibility(conditions):
            self.reward = 0.0
            for bf, weight in zip(self.basis_functions, self.basis_weights):
                basis_value = bf(conditions, indx=hpid)
                self.reward += basis_value * weight

            if not np.isscalar(self.reward):
                self.reward = np.sum(self.reward[hpid])

                if np.any(np.isinf(self.reward)):
                    self.reward = np.inf
        else:
            # If not feasible, negative infinity reward
            self.reward = -np.inf

        return self.reward



import unittest

import numpy as np

from rubin_scheduler.site_models import CloudModel


class TestCloudModel(unittest.TestCase):
    def test_call(self):
        cloud_model = CloudModel()
        in_cloud = 1.53
        efd_data = {"cloud": in_cloud}
        alt = np.zeros(50, float)
        az = np.zeros(50, float)
        target_dict = {"altitude": alt, "azimuth": az}
        out_cloud = cloud_model(efd_data, target_dict)["cloud"]
        # Test that we propagated cloud value over the whole sky.
        self.assertEqual(in_cloud, out_cloud.max())
        self.assertEqual(in_cloud, out_cloud.min())
        self.assertEqual(len(out_cloud), len(alt))


if __name__ == "__main__":
    unittest.main()



__all__ = ("Conditions",)

import functools
import warnings
from inspect import signature
from io import StringIO

import healpy as hp
import numpy as np
import pandas as pd

from rubin_scheduler.scheduler.utils import match_hp_resolution, smallest_signed_angle
from rubin_scheduler.utils import (
    DEFAULT_NSIDE,
    Site,
    _angular_separation,
    _approx_altaz2pa,
    _approx_ra_dec2_alt_az,
    _hpid2_ra_dec,
    calc_lmst,
    m5_flat_sed,
)


class Conditions:
    """Holds telemetry information, keeping calculated values in sync
    for `self.mjd` (such as ra/dec mappings to alt/az).

    Incoming values have setters to keep values in sync. Healpix maps
    are set to the expected (`self.nside`) resolution.

    Unless otherwise noted, all values are assumed to be valid at the time
    given by `self.mjd`.

    Parameters
    ----------
    nside : `int`, optional
        The healpixel nside to set the resolution of attributes.
        Default of None will use
        `rubin_scheduler.scheduler.utils.set_default_nside`.
    site : str ('LSST')
        A site name used to create a sims.utils.Site object. For
        looking up observatory parameters like latitude and longitude.
    exptime : `float`, optional
        The exposure time (seconds) to assume when computing the 5-sigma
        limiting depth maps stored in Conditions. These maps are used
        by basis functions, but are not used to calculate expected
        observation m5 values (such as when exposure time varies).
        Default 30 seconds.
    mjd : `float`, optional
        The current MJD.
        Default of None is fine on init - will be updated by telemetry
        stream.
    """

    global_maps = set(["ra", "dec", "slewtime", "airmass", "zeros_map", "nan_map"])
    by_band_maps = set(["skybrightness", "fwhm_eff", "m5_depth"])

    def __init__(
        self,
        nside=DEFAULT_NSIDE,
        site="LSST",
        exptime=30.0,
        mjd=None,
    ):
        """
        Attributes (Set on init)
        -----------
        nside : `int`
            Healpix resolution. All maps are set to this reslution.
        site : rubin_scheduler.Site object ('LSST')
            Contains static site-specific data (lat, lon, altitude, etc).
            Defaults to 'LSST'.
        ra : `np.ndarray`, (N,)
            A healpix array with the RA of each healpixel center (radians).
            Automatically generated.
        dec : `np.ndarray`, (N,)
            A healpix array with the Dec of each healpixel center (radians).
            Automatically generated.

        Attributes (to be set by user/telemetry stream/scheduler)
        -------------------------------------------
        mjd : `float`
            Modified Julian Date (days).
        bulk_cloud : `float`
            The fraction of sky covered by clouds. (In the future might
            update to transparency map)
        cloud_map : `np.ndarray`, (N,)
            XXX--to be done. HEALpix array with cloudy pixels set to NaN.
        slewtime : `np.ndarray`, (N,)
            Healpix showing the slewtime to each healpixel center (seconds)
        current_filter : `str`
            The name of the current filter. (expect one of u, g, r, i, z, y).
        mounted_filters : `list` [`str`]
            The filters that are currently mounted and thus available
            (expect 5 of u, g, r, i, z, y for LSSTCam).
        night : `int`
            The current night number (days). Probably starts at 1.
        skybrightness : `dict` {`str: `np.ndarray`, (N,)}
            Dictionary keyed by filter name.
            Values are healpix arrays with the sky brightness at each
            healpix center (mag/acsec^2)
        fwhm_eff : `dict` {`str: `np.ndarray`, (N,)}
            Dictionary keyed by filtername.
            Values are the effective seeing FWHM at each healpix
            center (arcseconds)
        moon_alt : `float`
            The altitude of the Moon (radians)
        moon_az : `float`
            The Azimuth of the moon (radians)
        moon_ra : `float`
            RA of the moon (radians)
        moon_dec : `float`
            Declination of the moon (radians)
        moon_phase : `float`
            The Phase of the moon. (percent, 0=new moon, 100=full moon)
        sun_alt : `float`
            The altitude of the sun (radians).
        sun_az : `float`
            The Azimuth of the sun (radians).
        sun_ra : `float`
            The RA of the sun (radians).
        sun_dec : `float`
            The Dec of the sun (radians).
        tel_ra : `float`
            The current telescope RA pointing (radians).
        tel_dec : `float`
            The current telescope Declination (radians).
        tel_alt : `float`
            The current telescope altitude (radians).
        tel_az : `float`
            The current telescope azimuth (radians).
        cumulative_azimuth_rad : `float`
            The cumulative telescope azimuth (radians).
            Used for tracking cable wrap.
        wind_speed : `float`
            Wind speed (m/s).
        wind_direction : `float`
            Direction from which the wind originates. A direction of
            0.0 degrees means the wind originates from the north and 90.0
            degrees from the east (radians).
        sunset : `float`
            The MJD of sunset that starts the current night. Note MJDs of
            sunset, moonset, twilight times, etc are from interpolations.
            This means the sun may actually be slightly above/below the
            horizon at the given sunset time.
        sun_n12_setting : `float`
            The MJD of when the sun is at -12 degrees altitude and
            setting during the current night. From interpolation.
        sun_n18_setting : `float`
            The MJD when the sun is at -18 degrees altitude and setting
            during the current night. From interpolation.
        sun_n18_rising : `float`
            The MJD when the sun is at -18 degrees altitude and rising
            during the current night. From interpolation.
        sun_n12_rising : `float`
            The MJD when the sun is at -12 degrees altitude and rising
            during the current night. From interpolation.
        sunrise : `float`
            The MJD of sunrise during the current night. From interpolation
        moonrise : `float`
            The MJD of moonrise during the current night. From interpolation.
        moonset : `float`
            The MJD of moon set during the current night. From interpolation.
        moon_phase_sunset : `float`
            The phase of the moon (0-100 illuminated) at sunset.
            Useful for setting which filters should be loaded.
        targets_of_opportunity : `list` [`rubin_scheduler.scheduler.targetoO`]
            targetoO objects.
        planet_positions : `dict` {`str`: `float`}
            Dictionary of planet name and coordinate e.g., 'venus_RA',
            'mars_dec'
        scheduled_observations : `np.ndarray`, (M,)
            A list of MJD times when there are scheduled observations.
            Defaults to empty array.
        sky_az_limits : `list` [[`float`, `float`]]
            A list of lists giving valid azimuth ranges. e.g.,
            [0, 2*np.pi] would mean all azimuth values are valid, while
            [[0, np.pi/2], [3*np.pi/2, 2*np.pi]] would mean anywhere in
            the south is invalid.  Radians.
        sky_alt_limits : `list` [[`float`, `float`]]
            A list of lists giving valid altitude ranges. Radians.
        tel_az_limits : `list` [`float`, `float`]
            A simple two-element list giving the valid azimuth ranges
            for the telescope movement. Radians.
        tel_alt_limits : `list` [`float`, `float`]
            A simple two-element list giving the valid altitude ranges
            for the telescope movement. Radians

        Attributes (calculated on demand and cached)
        ------------------------------------------
        alt : `np.ndarray`, (N,)
            Altitude of each healpixel (radians). Recalculated if mjd is
            updated. Uses fast approximate equation for converting
            RA,Dec to alt,az.
        az : `np.ndarray`, (N,)
            Azimuth of each healpixel (radians). Recalculated if mjd is
            updated. Uses fast approximate equation for converting RA,Dec
            to alt,az.
        airmass : `np.ndarray`, (N,)
            A healpix map with the airmass value of each healpixel. (unitless)
        pa : `np.ndarray`, (N,)
            The parallactic angle of each healpixel (radians). Recalculated
            if mjd is updated. Based on the fast approximate alt,az values.
        lmst : `float`
            The local mean sidereal time (hours). Updates is mjd is changed.
        m5_depth : `dict` {`str: `np.ndarray`, (N,)}
            the 5-sigma limiting depth healpix maps, keyed by filtername
            (mags). Will be recalculated if the skybrightness, seeing,
            or airmass are updated.
        HA : `np.ndarray`, (N,)
            Healpix map of the hour angle of each healpixel (radians).
            Runs from 0 to 2pi.
        az_to_sun : `np.ndarray`, (N,)
            Healpix map of the azimuthal distance to the sun for each
            healpixel (radians)
        az_to_anitsun : `np.ndarray`, (N,)
            Healpix map of the azimuthal distance to the anit-sun for each
            healpixel (radians)
        solar_elongation : `np.ndarray`, (N,)
            Healpix map of the solar elongation (angular distance to the sun)
            for each healpixel (radians)

        Attributes (set by the scheduler)
        -------------------------------
        queue : `list` [`observation` objects]
            The current queue of observations core_scheduler is waiting to
            execute.

        """
        self.nside = nside

        # The RA, Dec grid we are using
        hpids = np.arange(hp.nside2npix(self.nside))
        self.ra, self.dec = _hpid2_ra_dec(self.nside, hpids)

        self.site = Site(site)
        self.exptime = exptime

        # Generate an empty map so we can copy when we need a new map
        self.zeros_map = np.zeros(hp.nside2npix(self.nside), dtype=float)
        self.nan_map = np.zeros(hp.nside2npix(self.nside), dtype=float)
        self.nan_map.fill(np.nan)

        # Set other attributes to Nones
        self._init_attributes()
        self.mjd = mjd

    def _init_attributes(self):
        """Initialize or clear all the attributes"""

        # Modified Julian Date (day)
        self._mjd = None
        # Altitude and azimuth. Dict with degrees and radians
        self._alt = None
        self._az = None
        self._pa = None

        # The cloud level. Fraction, but could upgrade to transparency map
        self.clouds = None
        self._slewtime = None
        self.current_filter = None
        self.mounted_filters = None
        self.night = None
        self._lmst = None
        # Should be a dict with filtername keys
        self._skybrightness = {}
        self._fwhm_eff = {}
        self._m5_depth = None
        self._airmass = None

        self.wind_speed = None
        self.wind_direction = None

        # Upcoming scheduled observations
        self.scheduled_observations = np.array([], dtype=float)

        # Attribute to hold the current observing queue
        self.queue = None

        # Moon
        self.moon_alt = None
        self.moon_az = None
        self.moon_ra = None
        self.moon_dec = None
        self.moon_phase = None

        # Sun
        self.sun_alt = None
        self.sun_az = None
        self.sun_ra = None
        self.sun_dec = None

        # Almanac information
        self.sunset = None
        self.sun_n12_setting = None
        self.sun_n18_setting = None
        self.sun_n18_rising = None
        self.sun_n12_rising = None
        self.sunrise = None
        self.moonrise = None
        self.moonset = None
        self.moon_phase_sunset = None

        self.planet_positions = None

        # Current telescope pointing
        self.tel_ra = None
        self.tel_dec = None
        self.tel_alt = None
        self.tel_az = None
        self.cumulative_azimuth_rad = None

        # Sky coverage limits.
        # These should be in radians.
        self.sky_az_limits = None
        self.sky_alt_limits = None
        # Kinematic model (real slew) limits.
        self.tel_alt_limits = None
        self.tel_az_limits = None

        # Full sky cloud map
        self._cloud_map = None
        self._HA = None

        # XXX--document
        self.bulk_cloud = None

        self.rot_tel_pos = None

        self.targets_of_opportunity = None

        # Potential attributes that get computed
        self._solar_elongation = None
        self._az_to_sun = None
        self._az_to_antisun = None

    @property
    def lmst(self):
        if self._lmst is None:
            self._lmst = calc_lmst(self.mjd, self.site.longitude_rad)

        return self._lmst

    @lmst.setter
    def lmst(self, value):
        self._lmst = value
        self._HA = None

    @property
    def HA(self):
        if self._HA is None:
            self.calc_ha()
        return self._HA

    def calc_ha(self):
        self._HA = np.radians(self.lmst * 360.0 / 24.0) - self.ra
        self._HA[np.where(self._HA < 0)] += 2.0 * np.pi

    @property
    def cloud_map(self):
        return self._cloud_map

    @cloud_map.setter
    def cloud_map(self, value):
        self._cloud_map = match_hp_resolution(value, nside_out=self.nside)

    @property
    def slewtime(self):
        return self._slewtime

    @slewtime.setter
    def slewtime(self, value):
        # Using 0 for start of night
        if np.size(value) == 1:
            self._slewtime = value
        else:
            self._slewtime = match_hp_resolution(value, nside_out=self.nside)

    @property
    def airmass(self):
        if self._airmass is None:
            self.calc_airmass()
        return self._airmass

    def calc_airmass(self):
        self._airmass = np.zeros(self._alt.size, dtype=float)
        self._airmass.fill(np.nan)
        alt_limit = self.tel_alt_limits
        if alt_limit is None:
            alt_limit = 0
        else:
            alt_limit = np.min(alt_limit)
        good = np.where(self._alt > alt_limit)
        self._airmass[good] = 1.0 / np.cos(np.pi / 2.0 - self._alt[good])

    @property
    def pa(self):
        if self._pa is None:
            self.calc_pa()
        return self._pa

    def calc_pa(self):
        self._pa = _approx_altaz2pa(self.alt, self.az, self.site.latitude_rad)

    @property
    def alt(self):
        if self._alt is None:
            self.calc_alt_az()
        return self._alt

    @property
    def az(self):
        if self._az is None:
            self.calc_alt_az()
        return self._az

    def calc_alt_az(self):
        self._alt, self._az = _approx_ra_dec2_alt_az(
            self.ra,
            self.dec,
            self.site.latitude_rad,
            self.site.longitude_rad,
            self._mjd,
        )

    @functools.lru_cache(maxsize=10)
    def future_alt_az(self, mjd):
        """Compute the altitude and azimuth for a future time.

        Returns
        -------
        altitude : `np.array`
            The altutude of each healpix at MJD (radians)
        azimuth : : `np.array`
            The azimuth of each healpix at MJD (radians)
        """
        alt, az = _approx_ra_dec2_alt_az(
            self.ra,
            self.dec,
            self.site.latitude_rad,
            self.site.longitude_rad,
            mjd,
        )
        return alt, az

    @property
    def mjd(self):
        return self._mjd

    @mjd.setter
    def mjd(self, value):
        # If MJD is changed, everything else is no longer valid, so re-init
        if self.mjd is not None:
            warnings.warn("Changing MJD and resetting all attributes.")
        self._init_attributes()
        self._mjd = value

    @property
    def skybrightness(self):
        return self._skybrightness

    @skybrightness.setter
    def skybrightness(self, indict):
        for key in indict:
            self._skybrightness[key] = match_hp_resolution(indict[key], nside_out=self.nside)
        # If sky brightness changes, need to recalc M5 depth.
        self._m5_depth = None

    @property
    def fwhm_eff(self):
        return self._fwhm_eff

    @fwhm_eff.setter
    def fwhm_eff(self, indict):
        for key in indict:
            self._fwhm_eff[key] = match_hp_resolution(indict[key], nside_out=self.nside)
        self._m5_depth = None

    @property
    def m5_depth(self):
        if self._m5_depth is None:
            self.calc_m5_depth()
        return self._m5_depth

    def calc_m5_depth(self):
        self._m5_depth = {}
        for filtername in self._skybrightness:
            good = ~np.isnan(self._skybrightness[filtername])
            self._m5_depth[filtername] = self.nan_map.copy()
            self._m5_depth[filtername][good] = m5_flat_sed(
                filtername,
                self._skybrightness[filtername][good],
                self._fwhm_eff[filtername][good],
                self.exptime,
                self._airmass[good],
            )

    def calc_solar_elongation(self):
        self._solar_elongation = _angular_separation(self.ra, self.dec, self.sun_ra, self.sun_dec)

    @property
    def solar_elongation(self):
        if self._solar_elongation is None:
            self.calc_solar_elongation()
        return self._solar_elongation

    def calc_az_to_sun(self):
        self._az_to_sun = smallest_signed_angle(self.ra, self.sun_ra)

    def calc_az_to_antisun(self):
        self._az_to_antisun = smallest_signed_angle(self.ra + np.pi, self.sun_ra)

    @property
    def az_to_sun(self):
        if self._az_to_sun is None:
            self.calc_az_to_sun()
        return self._az_to_sun

    @property
    def az_to_antisun(self):
        if self._az_to_antisun is None:
            self.calc_az_to_antisun()
        return self._az_to_antisun

    def set_auxtel_info(
        self,
        mjd,
        slewtime,
        sun_n18_setting,
        sun_n18_rising,
        moon_alt,
        moon_az,
        tel_alt_limits,
        tel_az_limits,
        sky_alt_limits,
        sky_az_limits,
        wind_speed,
        wind_direction,
        **kwargs,
    ):
        """Method to set all the information we expect will be required by
        a standard auxtel scheduler. Extra attributes can be set via **kwargs.

        Parameters
        ----------
        mjd : `float`
            Modified Julian Date (days).
        slewtime : `np.ndarray`, (N,)
            Healpix showing the slewtime to each healpixel center (seconds)
        moon_alt : `float`
            The altitude of the Moon (radians)
        moon_az : `float`
            The Azimuth of the moon (radians)
        wind_speed : `float`
            Wind speed (m/s).
        wind_direction : `float`
            Direction from which the wind originates. A direction of
            0.0 degrees means the wind originates from the north and 90.0
            degrees from the east (radians).
        sun_n18_setting : `float`
            The MJD when the sun is at -18 degrees altitude and setting
            during the current night. From interpolation.
        sun_n18_rising : `float`
            The MJD when the sun is at -18 degrees altitude and rising
            during the current night. From interpolation.
        sky_az_limits : `list` [[`float`, `float`]]
            A list of lists giving valid azimuth ranges. e.g.,
            [0, 2*np.pi] would mean all azimuth values are valid, while
            [[0, np.pi/2], [3*np.pi/2, 2*np.pi]] would mean anywhere in
            the south is invalid.  Radians.
        sky_alt_limits : `list` [[`float`, `float`]]
            A list of lists giving valid altitude ranges. Radians.
        tel_az_limits : `list` [`float`, `float`]
            A simple two-element list giving the valid azimuth ranges
            for the telescope movement. Radians.
        tel_alt_limits : `list` [`float`, `float`]
            A simple two-element list giving the valid altitude ranges
            for the telescope movement. Radians
        """
        self._init_attributes()
        auxtel_args = signature(self.set_auxtel_info)
        loc = locals()
        for key in auxtel_args.parameters.keys():
            setattr(self, key, loc[key])

        potential_attrs = dir(self)
        for key in kwargs:
            if key not in potential_attrs:
                warnings.warn("Setting unexpected Conditions attribute %s" % key)
            setattr(self, key, kwargs[key])

    def set_maintel_info(
        self,
        mjd,
        slewtime,
        current_filter,
        mounted_filters,
        night,
        skybrightness,
        fwhm_eff,
        moon_alt,
        moon_az,
        moon_ra,
        moon_dec,
        moon_phase,
        sun_alt,
        sun_az,
        sun_ra,
        sun_dec,
        tel_ra,
        tel_dec,
        tel_alt,
        tel_az,
        wind_speed,
        wind_direction,
        sun_n12_setting,
        sun_n12_rising,
        sun_n18_setting,
        sun_n18_rising,
        moonrise,
        moonset,
        planet_positions,
        tel_alt_limits,
        tel_az_limits,
        sky_alt_limits,
        sky_az_limits,
        **kwargs,
    ):
        """Method to set all the information we expect will be required by
        a standard maintel scheduler. Extra attributes can be set via **kwargs.

        mjd : `float`
            Modified Julian Date (days).
        slewtime : `np.ndarray`, (N,)
            Healpix showing the slewtime to each healpixel center (seconds)
        current_filter : `str`
            The name of the current filter. (expect one of u, g, r, i, z, y).
        mounted_filters : `list` [`str`]
            The filters that are currently mounted and thus available
            (expect 5 of u, g, r, i, z, y for LSSTCam).
        night : `int`
            The current night number (days). Probably starts at 1.
        skybrightness : `dict` {`str: `np.ndarray`, (N,)}
            Dictionary keyed by filter name.
            Values are healpix arrays with the sky brightness at each
            healpix center (mag/acsec^2)
        fwhm_eff : `dict` {`str: `np.ndarray`, (N,)}
            Dictionary keyed by filtername.
            Values are the effective seeing FWHM at each healpix
            center (arcseconds)
        moon_alt : `float`
            The altitude of the Moon (radians)
        moon_az : `float`
            The Azimuth of the moon (radians)
        moon_ra : `float`
            RA of the moon (radians)
        moon_dec : `float`
            Declination of the moon (radians)
        moon_phase : `float`
            The Phase of the moon. (percent, 0=new moon, 100=full moon)
        sun_alt : `float`
            The altitude of the sun (radians).
        sun_az : `float`
            The Azimuth of the sun (radians).
        sun_ra : `float`
            The RA of the sun (radians).
        sun_dec : `float`
            The Dec of the sun (radians).
        tel_ra : `float`
            The current telescope RA pointing (radians).
        tel_dec : `float`
            The current telescope Declination (radians).
        tel_alt : `float`
            The current telescope altitude (radians).
        tel_az : `float`
            The current telescope azimuth (radians).
        wind_speed : `float`
            Wind speed (m/s).
        wind_direction : `float`
            Direction from which the wind originates. A direction of
            0.0 degrees means the wind originates from the north and 90.0
            degrees from the east (radians).
        sun_n12_setting : `float`
            The MJD of when the sun is at -12 degrees altitude and
            setting during the current night. From interpolation.
        sun_n18_setting : `float`
            The MJD when the sun is at -18 degrees altitude and setting
            during the current night. From interpolation.
        sun_n18_rising : `float`
            The MJD when the sun is at -18 degrees altitude and rising
            during the current night. From interpolation.
        sun_n12_rising : `float`
            The MJD when the sun is at -12 degrees altitude and rising
            during the current night. From interpolation.
        moonrise : `float`
            The MJD of moonrise during the current night. From interpolation.
        moonset : `float`
            The MJD of moon set during the current night. From interpolation.
        moon_phase_sunset : `float`
            The phase of the moon (0-100 illuminated) at sunset.
            Useful for setting which filters should be loaded.
        targets_of_opportunity : `list` [`rubin_scheduler.scheduler.targetoO`]
            targetoO objects.
        planet_positions : `dict` {`str`: `float`}
            Dictionary of planet name and coordinate e.g., 'venus_RA',
            'mars_dec'
        sky_az_limits : `list` [[`float`, `float`]]
            A list of lists giving valid azimuth ranges. e.g.,
            [0, 2*np.pi] would mean all azimuth values are valid, while
            [[0, np.pi/2], [3*np.pi/2, 2*np.pi]] would mean anywhere in
            the south is invalid.  Radians.
        sky_alt_limits : `list` [[`float`, `float`]]
            A list of lists giving valid altitude ranges. Radians.
        tel_az_limits : `list` [`float`, `float`]
            A simple two-element list giving the valid azimuth ranges
            for the telescope movement. Radians.
        tel_alt_limits : `list` [`float`, `float`]
            A simple two-element list giving the valid altitude ranges
            for the telescope movement. Radians
        """
        self._init_attributes()

        maintel_args = signature(self.set_maintel_info)
        loc = locals()
        for key in maintel_args.parameters.keys():
            setattr(self, key, loc[key])

        potential_attrs = dir(self)
        for key in kwargs:
            if key not in potential_attrs:
                warnings.warn("Setting unexpected Conditions attribute %s" % key)
            setattr(self, key, kwargs[key])

    def set_attrs(self, **kwargs):
        """Convience function for setting lots of attributes at once.
        See __init__ docstring for full list of potential attributes."""
        potential_attrs = dir(self)
        for key in kwargs:
            if key not in potential_attrs:
                warnings.warn("Setting unexpected Conditions attribute %s" % key)
            setattr(self, key, kwargs[key])

    def __repr__(self):
        return f"<{self.__class__.__name__} mjd='{self.mjd}' at {hex(id(self))}>"

    def __str__(self):
        # If dependencies of to_markdown are not installed, fall back on repr
        try:
            pd.DataFrame().to_markdown()
        except ImportError:
            return repr(self)

        if self.mjd is None:
            # Many elements of the pretty str representation rely on
            # the time beging set. If it is not, just return the ugly form.
            return repr(self)

        output = StringIO()
        print(f"{self.__class__.__qualname__} at {hex(id(self))}", file=output)
        print("============================", file=output)
        print("nside: ", self.nside, "  ", file=output)
        print("site: ", self.site.name, "  ", file=output)
        print("exptime: ", self.exptime, "  ", file=output)
        print("lmst: ", self.lmst, "  ", file=output)
        print("clouds: ", self.clouds, "  ", file=output)
        print("current_filter: ", self.current_filter, "  ", file=output)
        print("mounted_filters: ", self.mounted_filters, "  ", file=output)
        print("night: ", self.night, "  ", file=output)
        print("wind_speed: ", self.wind_speed, "  ", file=output)
        print("wind_direction: ", self.wind_direction, "  ", file=output)
        print(
            "len(scheduled_observations): ",
            len(self.scheduled_observations),
            "  ",
            file=output,
        )
        print(
            "len(queue): ",
            None if self.queue is None else len(self.queue),
            "  ",
            file=output,
        )
        print("moonPhase: ", self.moon_phase, "  ", file=output)
        print("bulk_cloud: ", self.bulk_cloud, "  ", file=output)
        print("targets_of_opportunity: ", self.targets_of_opportunity, "  ", file=output)
        print("cumulative_azimuth_rad: ", self.cumulative_azimuth_rad, "  ", file=output)

        positions = [
            {
                "name": "sun",
                "alt": self.sun_alt,
                "az": self.sun_az,
                "RA": self.sun_ra,
                "decl": self.sun_dec,
            }
        ]
        positions.append(
            {
                "name": "moon",
                "alt": self.moon_alt,
                "az": self.moon_az,
                "RA": self.moon_ra,
                "decl": self.moon_dec,
            }
        )
        for planet_name in ("venus", "mars", "jupiter", "saturn"):
            positions.append(
                {
                    "name": planet_name,
                    "RA": float(np.max(self.planet_positions[planet_name + "_RA"])),
                    "decl": float(np.max(self.planet_positions[planet_name + "_dec"])),
                }
            )
        positions.append(
            {
                "name": "telescope",
                "alt": np.max(self.tel_alt),
                "az": np.max(self.tel_az),
                "RA": np.max(self.tel_ra),
                "decl": np.max(self.tel_dec),
                "rot": np.max(self.rot_tel_pos),
            }
        )
        positions = pd.DataFrame(positions).set_index("name")
        print(file=output)
        print("Positions (radians)", file=output)
        print("-------------------", file=output)
        print(positions.to_markdown(), file=output)

        positions_deg = np.degrees(positions)
        print(file=output)
        print("Positions (degrees)", file=output)
        print("-------------------", file=output)
        print(positions_deg.to_markdown(), file=output)

        events = (
            "mjd",
            "sunset",
            "sun_n12_setting",
            "sun_n18_setting",
            "sun_n18_rising",
            "sun_n12_rising",
            "sunrise",
            "moonrise",
            "moonset",
            "sun_0_setting",
            "sun_0_rising",
        )
        event_rows = []
        for event in events:
            try:
                mjd = getattr(self, event)
                time = pd.to_datetime(float(mjd) + 2400000.5, unit="D", origin="julian")
                event_rows.append({"event": event, "MJD": mjd, "date": time})
            except AttributeError:
                pass

        event_df = pd.DataFrame(event_rows).set_index("event").sort_values(by="MJD")
        print("", file=output)
        print("Events", file=output)
        print("------", file=output)
        print(event_df.to_markdown(), file=output)

        map_stats = []
        for map_name in self.global_maps - set(["zeros_map", "nan_map"]):
            try:
                values = getattr(self, map_name)
                map_stats.append(
                    {
                        "map": map_name,
                        "nside": hp.npix2nside(len(values)),
                        "min": np.nanmin(values),
                        "max": np.nanmax(values),
                        "median": np.nanmedian(values),
                    }
                )
            except AttributeError:
                pass

        for base_map_name in self.by_band_maps:
            for band in "ugrizy":
                try:
                    values = getattr(self, base_map_name)[band]
                    map_name = f"{base_map_name}_{band}"
                    map_stats.append(
                        {
                            "map": map_name,
                            "nside": hp.npix2nside(len(values)),
                            "min": np.nanmin(values),
                            "max": np.nanmax(values),
                            "median": np.nanmedian(values),
                        }
                    )
                except AttributeError:
                    pass

        maps_df = pd.DataFrame(map_stats).set_index("map")
        print("", file=output)
        print("Maps", file=output)
        print("----", file=output)
        print(maps_df.to_markdown(), file=output)

        result = output.getvalue()
        return result

    def _repr_markdown_(self):
        return str(self)



__all__ = ("VaryExptDetailer", "calc_target_m5s")

import healpy as hp
import numpy as np
from scipy.stats import binned_statistic

from rubin_scheduler.scheduler.detailers import BaseDetailer
from rubin_scheduler.skybrightness_pre import dark_sky
from rubin_scheduler.utils import DEFAULT_NSIDE, Site, hpid2_ra_dec, m5_flat_sed, ra_dec2_hpid


def calc_target_m5s(alt=65.0, fiducial_seeing=0.9, exptime=20.0):
    """Use the skybrightness model to find some good target m5s.

    Parameters
    ----------
    alt : `float`, opt
        Altitude for the target, degrees. Default 65.
    fiducial_seeing : `float`, opt
        Fiducial FWHMeff seeing, arcseconds. Default 0.9.
    exptime : `float`, opt
        Exposure time for the comparison, seconds. Default 20.

    Returns
    -------
    goal_m5 : `dict` of `float`
        dictionary of expected m5 values keyed by filtername
    """

    nside = DEFAULT_NSIDE
    dark = dark_sky(nside=nside)
    hpid = np.arange(dark.size, dtype=int)
    ra, dec = hpid2_ra_dec(nside, hpid)
    site = Site(name="LSST")
    alts = site.latitude - dec + 90
    alts[np.where(alts > 90)] -= 90
    binsize = 5.0
    alt_bins = np.arange(0, 90 + binsize, binsize)
    alts_mid = (alt_bins[0:-1] + alt_bins[1:]) / 2
    sky_mags = {}
    high_alts = np.where(alts > 0)[0]
    for filtername in dark.dtype.names:
        sky_mags[filtername], _be, _binn = binned_statistic(
            alts[high_alts], dark[filtername][high_alts], bins=alt_bins, statistic="mean"
        )
        sky_mags[filtername] = np.interp(alt, alts_mid, sky_mags[filtername])

    airmass = 1.0 / np.cos(np.pi / 2.0 - np.radians(alt))

    goal_m5 = {}
    for filtername in sky_mags:
        goal_m5[filtername] = m5_flat_sed(filtername, sky_mags[filtername], fiducial_seeing, exptime, airmass)

    return goal_m5


class VaryExptDetailer(BaseDetailer):
    """Vary the exposure time on observations to try and keep each
    observation at uniform depth.

    Parameters
    ----------
    min_expt : `float` (20.)
        The minimum exposure time to use (seconds).
    max_expt : `float` (100.)
        The maximum exposure time to use
    target_m5 : `dict` (None)
        Dictionary with keys of filternames as str and target 5-sigma
        depth values as floats. If none, the target_m5s are set to a
        min_expt exposure at X=1.1 in dark time.

    """

    def __init__(self, nside=DEFAULT_NSIDE, min_expt=20.0, max_expt=100.0, target_m5=None):
        """"""
        # Dict to hold all the features we want to track
        self.survey_features = {}
        self.nside = nside
        self.min_exp = min_expt
        self.max_exp = max_expt
        if target_m5 is None:
            self.target_m5 = {
                "g": 24.381615425253738,
                "i": 23.41810142458083,
                "r": 23.964359143049755,
                "u": 22.978794343692783,
                "y": 21.755612950787068,
                "z": 22.80377793629767,
            }
        else:
            self.target_m5 = target_m5

    def __call__(self, observation_list, conditions):
        """
        Parameters
        ----------
        observation_list : `list` of observations
            The observations to detail.
        conditions : `rubin_scheduler.scheduler.conditions` object

        Returns
        -------
        List of observations.
        """
        obs_array = np.concatenate(observation_list)
        hpids = ra_dec2_hpid(self.nside, obs_array["RA"], obs_array["dec"])
        new_expts = np.zeros(obs_array.size, dtype=float)
        for filtername in np.unique(obs_array["filter"]):
            in_filt = np.where(obs_array["filter"] == filtername)
            delta_m5 = self.target_m5[filtername] - conditions.M5Depth[filtername][hpids[in_filt]]
            # We can get NaNs because dithering pushes the center of the
            # pointing into masked regions.
            nan_indices = np.argwhere(np.isnan(delta_m5)).ravel()
            for indx in nan_indices:
                bad_hp = hpids[in_filt][indx]
                # Note this might fail if we run at higher resolution,
                # then we'd need to look farther for pixels to interpolate.
                near_pix = hp.get_all_neighbours(conditions.nside, bad_hp)
                vals = conditions.M5Depth[filtername][near_pix]
                if True in np.isfinite(vals):
                    estimate_m5 = np.mean(vals[np.isfinite(vals)])
                    delta_m5[indx] = self.target_m5[filtername] - estimate_m5
                else:
                    raise ValueError("Failed to find a nearby unmasked sky value.")

            new_expts[in_filt] = conditions.exptime * 10 ** (delta_m5 / 1.25)
        new_expts = np.clip(new_expts, self.min_exp, self.max_exp)
        # I'm not sure what level of precision we can expect, so let's
        # just limit to seconds
        new_expts = np.round(new_expts)

        for i, observation in enumerate(observation_list):
            observation["exptime"] = new_expts[i]

        return observation_list



import unittest

import numpy as np

from rubin_scheduler.utils import LsstCameraFootprint


class TestLsstcamerafootprint(unittest.TestCase):
    def setUp(self):
        self.obj_ra = np.array([10.0, 12.1], float)
        self.obj_dec = np.array([-30.0, -30.0], float)
        self.obs_ra = np.array([10.0, 10.0], float)
        self.obs_dec = np.array([-30.0, -30.0], float)
        self.obs_rot_sky_pos = np.zeros(2)

    def test_camera(self):
        camera = LsstCameraFootprint(
            units="degrees",
        )
        idx_obs = camera(self.obj_ra, self.obj_dec, self.obs_ra, self.obs_dec, self.obs_rot_sky_pos)
        # The first of these objects should be in the middle of
        # the FOV, while the second is outside
        self.assertEqual(idx_obs, [0])


if __name__ == "__main__":
    unittest.main()



__all__ = [
    "rotation_converter",
    "RotationConverter",
    "RotationConverterComCam",
    "RotationConverterAuxtel",
    "pseudo_parallactic_angle",
]

import astropy.units as u
import numpy as np
from astropy.coordinates import AltAz, EarthLocation, SkyCoord
from astropy.time import Time


def pseudo_parallactic_angle(
    ra: float | np.ndarray,
    dec: float | np.ndarray,
    mjd: float | np.ndarray,
    lon: float = -70.7494,
    lat: float = -30.2444,
    height: float = 2650.0,
    pressure: float = 750.0,
    temperature: float = 11.5,
    relative_humidity: float = 0.4,
    obswl: float = 1.0,
):
    """Compute the pseudo parallactic angle.

    The (traditional) parallactic angle is the angle zenith - coord - NCP
    where NCP is the true-of-date north celestial pole.  This function instead
    computes zenith - coord - NCP_ICRF where NCP_ICRF is the north celestial
    pole in the International Celestial Reference Frame.
    See: https://smtn-019.lsst.io/v/DM-44258/index.html

    Parameters
    ----------
    ra, dec : `float`
        ICRF coordinates in degrees.
    mjd : `float`
        Modified Julian Date.
    latitude, longitude : float
        Geodetic coordinates of observer in degrees.
    height : `float`
        Height of observer above reference ellipsoid in meters.
    pressure : `float`
        Atmospheric pressure in millibars.
    temperature : `float`
        Atmospheric temperature in degrees Celsius.
    relative_humidity : `float`
    obswl : `float`
        Observation wavelength in microns.

    Returns
    -------
    ppa : float
        The pseudo parallactic angle in degrees.
    alt : float
        Altitude of the observations
    az : float
        Azimuth of the observations
    """
    obstime = Time(mjd, format="mjd", scale="tai")
    location = EarthLocation.from_geodetic(
        lon=lon * u.deg,
        lat=lat * u.deg,
        height=height * u.m,
        ellipsoid="WGS84",  # For concreteness
    )

    coord_kwargs = dict(
        obstime=obstime,
        location=location,
        pressure=pressure * u.mbar,
        temperature=temperature * u.deg_C,
        relative_humidity=relative_humidity,
        obswl=obswl * u.micron,
    )

    coord = SkyCoord(ra * u.deg, dec * u.deg, **coord_kwargs)

    towards_zenith = SkyCoord(
        alt=coord.altaz.alt + 10 * u.arcsec, az=coord.altaz.az, frame=AltAz, **coord_kwargs
    )

    towards_north = SkyCoord(ra=coord.icrs.ra, dec=coord.icrs.dec + 10 * u.arcsec, **coord_kwargs)

    ppa = coord.position_angle(towards_zenith) - coord.position_angle(towards_north)
    return ppa.wrap_at(180 * u.deg).deg, coord.altaz.alt.deg, coord.altaz.az.deg


def rotation_converter(telescope="rubin"):
    """Return the correct RotationConverter object."""
    if telescope.lower() == "rubin":
        return RotationConverter()
    elif telescope.lower() == "auxtel":
        return RotationConverterAuxtel()
    elif telescope.lower() == "comcam":
        return RotationConverterComCam()
    else:
        raise ValueError("Unknown telescope name")


def _wrap_180(in_angle):
    """Convert angle(s) to run from -180 to 180

    Parameters
    ----------
    in_angle : `float`
        Input angle in radians.
    """
    angle = in_angle % (2.0 * np.pi)
    if np.size(angle) == 1:
        if angle > np.pi:
            result = angle - 2.0 * np.pi
            return result
        else:
            return angle
    else:
        indx = np.where(angle > np.pi)[0]
        angle[indx] = angle[indx] - 2.0 * np.pi
        return angle


class RotationConverter(object):
    """Class to convert between rotTelPos and rotSkyPos for LSSTcam."""

    def rottelpos2rotskypos(self, rottelpos_in, pa):
        """convert rotTelPos to rotSkyPos

        Parameters
        ----------
        rottelpos_in : `float`
            RotTelPos value in degrees
        pa : `float`
            Parallactic angle in degrees.
        """
        return np.degrees(self._rottelpos2rotskypos(np.radians(rottelpos_in), np.radians(pa)))

    def rotskypos2rottelpos(self, rotskypos_in, pa):
        """convert rotSkyPos to rotTelPos

        Parameters
        ----------
        rotskypos_in : `float`
            RotSkyPos value in degrees
        pa : `float`
            Parallactic angle in degrees.
        """
        result = self._rotskypos2rottelpos(np.radians(rotskypos_in), np.radians(pa))
        return np.degrees(result)

    def _rottelpos2rotskypos(self, rottelpos_in, pa):
        result = (pa - rottelpos_in - np.pi / 2) % (2.0 * np.pi)
        return result

    def _rotskypos2rottelpos(self, rotskypos_in, pa):
        result = (pa - rotskypos_in - np.pi / 2) % (2.0 * np.pi)
        # Enforce rotTelPos between -pi and pi
        return _wrap_180(result)


class RotationConverterComCam(RotationConverter):
    """Class to convert between rotTelPos and rotSkyPos for ComCam."""

    def _rottelpos2rotskypos(self, rottelpos_in, pa):
        result = (pa - rottelpos_in) % (2.0 * np.pi)
        return result

    def _rotskypos2rottelpos(self, rotskypos_in, pa):
        result = (pa - rotskypos_in) % (2.0 * np.pi)
        # Enforce rotTelPos between -pi and pi
        return _wrap_180(result)


class RotationConverterAuxtel(RotationConverter):
    """Use a different relation for rotation angles on AuxTel"""

    def _rottelpos2rotskypos(self, rottelpos_in, pa):
        return (rottelpos_in - pa) % (2.0 * np.pi)

    def _rotskypos2rottelpos(self, rotskypos_in, pa):
        result = (rotskypos_in + pa) % (2.0 * np.pi)
        # Enforce rotTelPos between -pi and pi
        return _wrap_180(result)



import gzip
import os
import pickle
import unittest
from glob import glob
from math import ceil
from pathlib import Path
from tempfile import TemporaryDirectory

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.scheduler.example import example_scheduler, run_sched
from rubin_scheduler.scheduler.features import Conditions
from rubin_scheduler.scheduler.schedulers import CoreScheduler
from rubin_scheduler.utils import SURVEY_START_MJD


class TestSnapshots(unittest.TestCase):

    @unittest.skip("Test too slow to run routinely.")
    @unittest.skipUnless(
        os.path.isfile(os.path.join(get_data_dir(), "scheduler/dust_maps/dust_nside_32.npz")),
        "Test data not available.",
    )
    def test_snapshots(self):
        """Test the saving of scheduler snapshots"""
        scheduler = example_scheduler(mjd_start=SURVEY_START_MJD, nside=8)

        # Find a start time definitely in the middle of the night
        mjd_start = ceil(SURVEY_START_MJD) + 0.15

        # only run for a few minutes
        survey_length = 4.0 / (60 * 60)

        with TemporaryDirectory() as test_dir:
            run_sched(
                scheduler,
                mjd_start=mjd_start,
                survey_length=survey_length,
                nside=scheduler.nside,
                snapshot_dir=test_dir,
            )
            snapshot_files = glob("sched_snapshot_*.p.gz", root_dir=test_dir)
            assert len(snapshot_files) > 0
            test_snapshot = Path(test_dir).joinpath(snapshot_files[0])
            with gzip.open(test_snapshot, "rb") as pio:
                scheduler_snapshot, conditions_snapshot = pickle.load(pio)
            assert isinstance(scheduler_snapshot, CoreScheduler)
            assert isinstance(conditions_snapshot, Conditions)



import unittest

import healpy as hp
import numpy as np
import pandas as pd

import rubin_scheduler.scheduler.basis_functions as basis_functions
import rubin_scheduler.scheduler.surveys as surveys
from rubin_scheduler.scheduler.basis_functions import SimpleArrayBasisFunction
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.schedulers import CoreScheduler
from rubin_scheduler.scheduler.utils import HpInLsstFov, ObservationArray, ScheduledObservationArray
from rubin_scheduler.utils import DEFAULT_NSIDE, SURVEY_START_MJD


def make_observations_list(nobs=1):
    observations_list = []
    for i in range(0, nobs):
        observation = ObservationArray()
        observation["mjd"] = SURVEY_START_MJD + i * 30 / 60 / 60 / 24
        observation["RA"] = np.radians(30)
        observation["dec"] = np.radians(-20)
        observation["filter"] = "r"
        observation["scheduler_note"] = "test"
        observations_list.append(observation)
    return observations_list


def make_observations_arrays(observations_list, nside=32):
    # Turn list of observations (that should already have useful info)
    # into observations_array plus observations_hpids_array.
    observations_array = np.empty(len(observations_list), dtype=observations_list[0].dtype)
    for i, obs in enumerate(observations_list):
        observations_array[i] = obs
    # Build observations_hpids_array.
    # Find list of lists of healpixels
    # (should match [indxs, indxs, indxs2] from above)
    pointing2indx = HpInLsstFov(nside=nside)
    list_of_hpids = pointing2indx(observations_array["RA"], observations_array["dec"])
    # Unravel list-of-lists (list_of_hpids) to match against observations
    hpids = []
    big_array_indx = []
    for i, indxs in enumerate(list_of_hpids):
        for indx in indxs:
            hpids.append(indx)
            big_array_indx.append(i)
    hpids = np.array(hpids, dtype=[("hpid", int)])
    # Set up format / dtype for observations_hpids_array
    names = list(observations_array.dtype.names)
    types = [observations_array[name].dtype for name in names]
    names.append(hpids.dtype.names[0])
    types.append(hpids["hpid"].dtype)
    ndt = list(zip(names, types))
    observations_hpids_array = np.empty(hpids.size, dtype=ndt)
    # Populate observations_hpid_array - big_array_indx points
    # between index in observations_array and index in hpid
    observations_hpids_array[list(observations_array.dtype.names)] = observations_array[big_array_indx]
    observations_hpids_array[hpids.dtype.names[0]] = hpids
    return observations_array, observations_hpids_array


class TestSurveys(unittest.TestCase):
    def test_field_survey(self):
        nside = 32

        bfs = []
        bfs.append(basis_functions.M5DiffBasisFunction(nside=nside))

        survey = surveys.FieldSurvey(bfs, RA=90.0, dec=-30.0)

        observatory = ModelObservatory()

        # Check dunder methods
        self.assertIsInstance(repr(survey), str)
        self.assertIsInstance(str(survey), str)

        # Check survey access methods
        conditions = observatory.return_conditions()
        reward = survey.calc_reward_function(conditions)
        self.assertIsInstance(reward, float)
        reward_df = survey.make_reward_df(conditions)
        self.assertIsInstance(reward_df, pd.DataFrame)
        _ = survey.make_reward_df(conditions, accum=False)

    def test_field_survey_add_observations(self):
        nside = 32
        # Just need a placeholder for this.
        bfs = [basis_functions.M5DiffBasisFunction(nside=nside)]

        observations_list = make_observations_list(10)
        indexes = []
        pointing2hpindx = HpInLsstFov(nside=nside)
        for i, obs in enumerate(observations_list):
            obs["mjd"] = SURVEY_START_MJD + i
            obs["rotSkyPos"] = 0
            if i < 5:
                obs["filter"] = "r"
                obs["scheduler_note"] = "r band"
            else:
                obs["filter"] = "g"
                obs["scheduler_note"] = "g band"
            obs["RA"] = np.radians(90)
            obs["dec"] = np.radians(-30)
            indexes.append(pointing2hpindx(obs["RA"], obs["dec"], rotSkyPos=obs["rotSkyPos"]))
        observations_array, observations_hpid_array = make_observations_arrays(observations_list)

        # Try adding observations to survey one at a time.
        survey = surveys.FieldSurvey(bfs, RA=90.0, dec=-30.0)
        for obs, indx in zip(observations_list, indexes):
            survey.add_observation(obs, indx=indx)
        self.assertTrue(survey.extra_features["ObsRecorded"].feature == len(observations_list))
        self.assertTrue(survey.extra_features["LastObs"].feature["mjd"] == observations_list[-1]["mjd"])
        # Try adding observations to survey in array
        survey = surveys.FieldSurvey(bfs, RA=90.0, dec=-30.0)
        survey.add_observations_array(observations_array, observations_hpid_array)
        self.assertTrue(survey.extra_features["ObsRecorded"].feature == len(observations_list))
        self.assertTrue(survey.extra_features["LastObs"].feature["mjd"] == observations_list[-1]["mjd"])

        # Now with different scheduler_notes.
        survey = surveys.FieldSurvey(bfs, RA=90.0, dec=-30.0, scheduler_note="r band")
        for obs, indx in zip(observations_list, indexes):
            survey.add_observation(obs, indx=indx)
        self.assertTrue(survey.extra_features["ObsRecorded"].feature == len(observations_list))
        self.assertTrue(survey.extra_features["LastObs"].feature["mjd"] == observations_list[-1]["mjd"])
        self.assertTrue(survey.extra_features["ObsRecorded_note"].feature == 5)
        self.assertTrue(survey.extra_features["LastObs_note"].feature["mjd"] == observations_list[4]["mjd"])
        # Try adding observations to survey in array
        survey = surveys.FieldSurvey(bfs, RA=90.0, dec=-30.0, scheduler_note="r band")
        survey.add_observations_array(observations_array, observations_hpid_array)
        self.assertTrue(survey.extra_features["ObsRecorded"].feature == len(observations_list))
        self.assertTrue(survey.extra_features["LastObs"].feature["mjd"] == observations_list[-1]["mjd"])
        self.assertTrue(survey.extra_features["ObsRecorded_note"].feature == 5)
        self.assertTrue(survey.extra_features["LastObs_note"].feature["mjd"] == observations_list[4]["mjd"])

    def test_field_altaz(self):
        """Test that we can use alt,az pointings"""
        survey = surveys.FieldAltAzSurvey([], alt=75, az=180.0)
        observatory = ModelObservatory()
        conditions = observatory.return_conditions()

        reward = survey.calc_reward_function(conditions)
        self.assertIsInstance(reward, float)
        reward_df = survey.make_reward_df(conditions)
        self.assertIsInstance(reward_df, pd.DataFrame)
        _ = survey.make_reward_df(conditions, accum=False)

        observation = survey.generate_observations(conditions)
        assert observation[0].dtype == ObservationArray().dtype

        # Advance time slightly
        observatory.mjd = observatory.mjd + 0.1
        conditions = observatory.return_conditions()

        observation_2 = survey.generate_observations(conditions)

        # Dec should stay the same, but RA will shift
        assert observation[0]["dec"] == observation_2[0]["dec"]
        assert observation[0]["RA"] != observation_2[0]["RA"]

        # Test that masks are respected
        # Point at a masked altitude. Reward should be nan or -inf
        survey = surveys.FieldAltAzSurvey(
            [basis_functions.AltAzShadowMaskBasisFunction(max_alt=85.0)], alt=90, az=180.0
        )
        reward = survey.calc_reward_function(conditions)
        assert ~np.isfinite(reward)

        # Point at an allowed region. Reward should be finite now.
        survey = surveys.FieldAltAzSurvey(
            [basis_functions.AltAzShadowMaskBasisFunction(max_alt=85.0)], alt=80, az=180.0
        )
        reward = survey.calc_reward_function(conditions)
        assert np.isfinite(reward)

    def test_scripted_survey(self):
        """Test the add observations methods of scripted surveys"""

        survey = surveys.ScriptedSurvey([])

        observations = ScheduledObservationArray(n=5)
        observations["filter"] = "r"
        observations["scheduler_note"] = ["a", "a", "c", "d", "a"]
        observations["mjd_tol"] = 1
        observations["mjd"] = 1
        observations["flush_by_mjd"] = 2

        survey.set_script(observations)

        # Make a completed list that has the first 3 observations
        completed_observations = ObservationArray(n=3)
        completed_observations["filter"] = survey.obs_wanted["filter"][0:3]
        completed_observations["scheduler_note"] = survey.obs_wanted["scheduler_note"][0:3]

        # Add one at a time
        for obs in completed_observations:
            not_a_slice = ObservationArray(n=1)
            for key in obs.dtype.names:
                not_a_slice[key] = obs[key]
            survey.add_observation(not_a_slice)

        assert np.sum(survey.obs_wanted["observed"]) == 3

        # Init to clear the internal counter
        survey = surveys.ScriptedSurvey([])
        observations["scheduler_note"] = ["a", "a", "c", "d", "a"]
        survey.set_script(observations)
        sched = CoreScheduler([survey])
        # Add full array at once
        sched.add_observations_array(completed_observations)
        assert np.sum(survey.obs_wanted["observed"]) == 3

        # Make sure error gets raised if we set a script wrong
        survey = surveys.ScriptedSurvey([])
        observations["scheduler_note"] = ["a", "a", "c", "d", "a"]
        with self.assertRaises(Exception) as context:
            survey.set_script(observations, add_index=False)
            self.assertTrue("unique scheduler_note" in str(context))

    def test_pointings_survey(self):
        """Test the pointing survey."""
        mo = ModelObservatory()
        conditions = mo.return_conditions()

        # Make a ring of points near the equator so
        # some should always be visible
        fields = ObservationArray(n=10)
        fields["RA"] = np.arange(0, fields.size) / fields.size * 2.0 * np.pi
        fields["dec"] = -0.01
        fields["scheduler_note"] = ["test%i" % ind for ind in range(fields.size)]
        fields["filter"] = "r"
        survey = surveys.PointingsSurvey(fields)

        reward = survey.calc_reward_function(conditions)
        assert np.isfinite(reward)

        obs = survey.generate_observations(conditions)
        # Confirm that our desired input values got passed through
        assert obs[0]["dec"] < 0
        assert obs[0]["scheduler_note"][0][0:4] == "test"

        # Adding observations
        assert np.sum(survey.n_obs) == 0
        survey.add_observation(obs[0])
        assert np.sum(survey.n_obs) == 1
        survey.add_observations_array(fields, None)
        assert np.sum(survey.n_obs) == 11

        # Check we can get display things out
        rc = survey.reward_changes(conditions)
        assert len(rc) == len(survey.weights)

        # Check we get a dataFrame
        df = survey.make_reward_df(conditions)
        assert len(df) == len(survey.weights)

    def test_roi(self):
        random_seed = 6563
        infeasible_hpix = 123
        nside = DEFAULT_NSIDE
        npix = hp.nside2npix(nside)
        rng = np.random.default_rng(seed=random_seed)
        num_bfs = 3
        bf_values = rng.random((num_bfs, npix))
        bf_values[:, infeasible_hpix] = -np.inf
        bfs = [SimpleArrayBasisFunction(values) for values in bf_values]

        observatory = ModelObservatory()
        conditions = observatory.return_conditions()

        # A few cases with an ROI with one valid healpix
        for i in range(3):
            hpix = rng.integers(npix)
            ra, decl = hp.pix2ang(nside, hpix, lonlat=True)
            survey = surveys.FieldSurvey(bfs, RA=ra, dec=decl)
            reward_df = survey.make_reward_df(conditions)
            for value, max_basis_reward in zip(bf_values[:, hpix], reward_df["max_basis_reward"]):
                self.assertEqual(max_basis_reward, value)

        # One case with an ROI with only an infeasible healpix
        ra, decl = hp.pix2ang(nside, infeasible_hpix, lonlat=True)
        survey = surveys.FieldSurvey(bfs, RA=ra, dec=decl)
        reward_df = survey.make_reward_df(conditions)
        for max_basis_reward in reward_df["max_basis_reward"]:
            self.assertEqual(max_basis_reward, -np.inf)

        for area in reward_df["basis_area"]:
            self.assertEqual(area, 0.0)

        for feasible in reward_df["feasible"]:
            self.assertFalse(feasible)

        # Make sure it still works as expected if no ROI is set
        weights = [1] * num_bfs
        survey = surveys.BaseMarkovSurvey(bfs, weights)
        reward_df = survey.make_reward_df(conditions)
        for value, max_basis_reward in zip(bf_values.max(axis=1), reward_df["max_basis_reward"]):
            self.assertEqual(max_basis_reward, value)


if __name__ == "__main__":
    unittest.main()



__all__ = ("calc_season",)

import numpy as np
from astropy.coordinates import EarthLocation, get_sun
from astropy.time import Time

from .constants import SURVEY_START_MJD


def calc_season(ra, mjd, mjd_start=None, calc_diagonal=False):
    """Calculate the season (of visibility) in the survey
    for a series of ra/time values of an observation.

    Based only on the RA of the point on the
    sky, it calculates the 'season' based on when the sun
    passes through this RA (this marks the start of a 'season').


    Parameters
    ----------
    ra : `float` or `np.ndarray` (N,)
        The RA (in degrees) of the point(s) on the sky
    mjd : `np.ndarray`, (M,)
        The times of the observations, in MJD days
    mjd_start : `float`, optional
        The time of the start of the survey.
        Default None will use `rubin_scheduler.utils.survey_start_mjd()`.
    calc_diagonal : `bool`, optional
        If True AND if len(mjd) == len(ra), ONLY the 'diagonal'
        elements of the seasons will be calculated.
        That is: RA and mjd are assumed to be paired and the resulting
        seasons will be a  1-d `np.array` of dimension (N,).

    Returns
    -------
    seasons : `np.array`, (M,) or `np.array` (N, M)
        The season values, as either a 1-d or 2-d array, depending on
        the value passed for ra and calc_diagonal.


    Notes
    -----
    Seasons should be calculated using the RA of a fixed point
    on the sky, such as the slice_point['ra'] if calculating season
    values for a series of opsim pointings on the sky, rather
    than the value for each visit.

    To convert to integer seasons, use np.floor(seasons).

    "Season 0" will be the first (full) season after the RA value becomes
    visible for the first time after mjd_start.
    This induces a "discontinuity" in the resulting season values,
    corresponding to the RA where the survey started -- like the international
    dateline (in this case, where season goes from
    'past season' to 'new season').

    The scheduler `season_calc` routine computes "season" independently,
    and may have a 90 degree offset in the location of this 'skip'/zeropoint,
    plus requires different calculation of `offset` values to account for
    location on the sky.
    """
    # Just to make it simpler to get simple values back
    try:
        len(ra)
        single_ra = False
    except TypeError:
        single_ra = True

    if mjd_start is None:
        mjd_start = SURVEY_START_MJD

    # A reference time and sun RA location to anchor the location of the Sun
    # This time was chosen as it is close to the expected start of the survey.
    # The time is an equinox point, so the sun is at dec=0.
    ref_time = 60940.0
    ref_sun_ra = 178.98403541421598

    # Calculate the fraction of the sphere/"year" for these RAs on the sky.
    offset = (ra - ref_sun_ra) / 360 * 365.25
    # Turn the offsets into the (mjd) times that the season would start
    # at each RA value, in the reference year.
    season_began = ref_time + offset

    # Add an adjustment so that the first season (season = 0)
    # at each RA is in the year after mjd_start
    first = np.floor((season_began - mjd_start) / 365.25)
    season_began = season_began - first * 365.25

    # Calculate the season value for each point.
    # season_began is an N length array, matching "ra"
    # mjds is an M length array, matching mjds (applied to every ra)
    if single_ra:
        seasons = (mjd - season_began) / 365.25
    elif calc_diagonal and len(mjd) == len(ra):
        seasons = (mjd - season_began) / 365.25
    else:
        seasons = np.zeros((len(ra), len(mjd)), dtype=float)
        for i, r in enumerate(ra):
            seasons[i] = (mjd - season_began[i]) / 365.25

    return seasons


def _generate_reference():
    # The reference values for calc_season can be evaluated using
    loc = EarthLocation.of_site("Rubin", refresh_cache=True)
    # time should ideally be near at equinox for RA projection considerations
    t = Time("2025-09-22T00:00:00.00", format="isot", scale="utc", location=loc)
    print("Ref time", t.utc.mjd)
    print("Ref sun RA", get_sun(t).ra.deg, t.utc.mjd)
    print("local sidereal time at time", t.sidereal_time("apparent").deg)



__all__ = ("DEFAULT_NSIDE", "SURVEY_START_MJD", "survey_start_mjd")

import warnings

DEFAULT_NSIDE = 32  # HEALpix nside, ~1.83 degree resolution
SURVEY_START_MJD = 60796.0  # 60796 = May 10, 2025


def survey_start_mjd():
    """For backwards compatibility"""
    warnings.warn("Function survey_start_mjd is deprecated, use SURVEY_START_MJD.", FutureWarning)

    return SURVEY_START_MJD



__all__ = ("sample_patch_on_sphere", "uniform_sphere")


import numpy as np


def uniform_sphere(npoints, seed=42):
    """
    Just make RA, dec points on a sphere
    """
    np.random.seed(seed)
    u = np.random.uniform(size=npoints)
    v = np.random.uniform(size=npoints)

    ra = 2.0 * np.pi * u
    dec = np.arccos(2.0 * v - 1.0)
    # astro convention of -90 to 90
    dec -= np.pi / 2.0
    return np.degrees(ra), np.degrees(dec)


def sample_patch_on_sphere(phi, theta, delta, size, seed=1):
    """
    Uniformly distributes samples on a patch on a sphere between phi pm delta,
    and theta pm delta on a sphere. Uniform distribution implies that the
    number of points in a patch of sphere is proportional to the area of the
    patch. Here, the coordinate system is the usual
    spherical coordinate system but with the azimuthal angle theta going from
    90 degrees at the North Pole, to -90 degrees at the South Pole, through
    0. at the equator.

    This function is not equipped to handle wrap-around the ranges of theta
    phi and therefore does not work at the poles.

    Parameters
    ----------
    phi : `float`,
        center of the spherical patch in ra with range, degrees
    theta : `float`
        degrees
    delta : `float`
        degrees
    size : `int`
        number of samples
    seed : `int`, optional, defaults to 1
        random Seed used for generating values

    Returns
    -------
    phivals, thetavals : `np.ndarray`, `np.ndarray`
        tuple of (phivals, thetavals) where phivals and thetavals are arrays of
        size size in degrees.
    """
    np.random.seed(seed)
    u = np.random.uniform(size=size)
    v = np.random.uniform(size=size)

    phi = np.radians(phi)
    theta = np.radians(theta)
    delta = np.radians(delta)

    phivals = 2.0 * delta * u + (phi - delta)
    phivals = np.where(phivals >= 0.0, phivals, phivals + 2.0 * np.pi)

    # use conventions in spherical coordinates
    theta = np.pi / 2.0 - theta

    thetamax = theta + delta
    thetamin = theta - delta

    if thetamax > np.pi or thetamin < 0.0:
        raise ValueError("Function not implemented to cover wrap around poles")

    # Cumulative Density Function is cos(thetamin) - cos(theta) /
    # cos(thetamin) - cos(thetamax)
    a = np.cos(thetamin) - np.cos(thetamax)
    thetavals = np.arccos(-v * a + np.cos(thetamin))

    # Get back to -pi/2 to pi/2 range of decs
    thetavals = np.pi / 2.0 - thetavals
    return np.degrees(phivals), np.degrees(thetavals)



import astropy.units as u
import numpy as np
from astroplan import Observer
from astropy.time import Time

from rubin_scheduler.utils import Site

if __name__ == "__main__":
    # Trying out the astroplan sunrise/set code.
    # conda install -c astropy astroplan
    mjd_start = 59853.5 - 3.0 * 365.25
    duration = 25.0 * 365.25
    pad_around = 40
    t_step = 0.7

    mjds = np.arange(mjd_start - pad_around, duration + mjd_start + pad_around + t_step, t_step)

    site = Site("LSST")
    observer = Observer(
        longitude=site.longitude * u.deg,
        latitude=site.latitude * u.deg,
        elevation=site.height * u.m,
        name="LSST",
    )

    # This blows up if I try to do it all at once? 250 GB of memory?
    results = []

    mjds_list = np.array_split(mjds, 500)

    for i, mjds in enumerate(mjds_list):
        print("chunk %i of %i" % (i, len(mjds_list)))
        times = Time(mjds, format="mjd")
        print("getting sunsets")
        sunsets = observer.sun_set_time(times)

        sunsets = np.unique(np.round(sunsets.mjd, decimals=4))

        names = [
            "night",
            "sunset",
            "sun_n12_setting",
            "sun_n18_setting",
            "sun_n18_rising",
            "sun_n12_rising",
            "sunrise",
            "moonrise",
            "moonset",
        ]
        types = [int]
        types.extend([float] * (len(names) - 1))
        almanac = np.zeros(sunsets.size, dtype=list(zip(names, types)))
        almanac["sunset"] = sunsets

        times = Time(sunsets, format="mjd")
        print("evening twilight 1")
        almanac["sun_n12_setting"] = observer.twilight_evening_nautical(times).mjd
        almanac["sun_n18_setting"] = observer.twilight_evening_astronomical(times).mjd
        almanac["sun_n18_rising"] = observer.twilight_morning_astronomical(times).mjd
        almanac["sun_n12_rising"] = observer.twilight_morning_nautical(times).mjd
        almanac["sunrise"] = observer.sun_rise_time(times).mjd
        almanac["moonset"] = observer.moon_set_time(times).mjd
        print("moonrise")
        almanac["moonrise"] = observer.moon_rise_time(times).mjd
        results.append(almanac)

    almanac = np.concatenate(results)
    umjds, indx = np.unique(almanac["sunset"], return_index=True)
    almanac = almanac[indx]
    almanac["night"] = np.arange(almanac["night"].size)

    np.savez("sunsets.npz", almanac=almanac)



import unittest

from rubin_scheduler.utils import m5_flat_sed


class PhotoM5Test(unittest.TestCase):
    def testm5(self):
        filters = ["u", "g", "r", "i", "z", "y"]
        kwargs = {}
        # List all parameters to test, with better conditions first
        kwargs["musky"] = [23.0, 22.0]
        kwargs["fwhm_eff"] = [1.0, 1.5]
        kwargs["exp_time"] = [60.0, 30.0]
        kwargs["airmass"] = [1.0, 2.2]
        kwargs["tau_cloud"] = [0.0, 2.2]

        k_default = {}
        for key in kwargs:
            k_default[key] = kwargs[key][0]

        for filtername in filters:
            m5_baseline = m5_flat_sed(filtername, **k_default)
            for key in kwargs:
                k_new = k_default.copy()
                k_new[key] = kwargs[key][1]
                m5_new = m5_flat_sed(filtername, **k_new)
                assert m5_new < m5_baseline


if __name__ == "__main__":
    unittest.main()
